import logging
import re
import requests

import pandas as pd


#
# Requests
#


def get_json(url, headers={}):
    response = requests.get(url, headers=headers)
    logging.info(f"{response.url}.. {response.status_code}")

    jo = {}
    if response.status_code == 200:
        jo = response.json()

    return jo


def camel_case_keys(df):
    last_dim = len(df.shape) - 1
    df = df.rename(
        {k: camel_to_snake_case(k) for k in df.keys()}, axis=last_dim
    ).rename(
        {
            "forward_pe1_yr": "forward_pe_1_yr",
            "is_nasdaq100": "is_nasdaq_100",
        },
        axis=last_dim,
    )
    return df


def camel_to_snake_case(name):
    name = re.sub(" ", "", name)
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()
    return name


#
# Pandas
#


def reindex(df: pd.DataFrame, moved: list, position: str = "start", axis: int = 1):
    assert axis == 1

    rest = [c for c in df if c not in moved]
    if position == "start":
        new_labels = moved + rest
    elif position == "end":
        new_labels = rest + moved
    else:
        # can support int position too
        raise ValueError()

    df_reindexed = df.reindex(new_labels, axis=axis)

    return df_reindexed


if False:
    camel_to_snake_case("getHTTPResponseCode") == "get_http_response_code"
    camel_to_snake_case("ForwardPE1Yr") == "forward_pe1_yr"
    camel_to_snake_case("Security Name") == "security_name"



import plotly.express as px

from .common import defaults


def plot_status(status, y='gain_pct',
                template=defaults['template'], **kwargs):

    total_cost, total_value = status.sum()[['cost', 'value']]
    title = f'My Holdings; Gain of ${total_value-total_cost:,.2f} ({total_value/total_cost-1:,.2%})'

    fig = px.bar(status.sort_values(y, ascending=False),
                 x='symbol', y=y, color=y,
                 text=y, hover_data=['gain'],
                 title=title,
                 template=template, color_continuous_midpoint=0,
                 **kwargs)
    fig.update_layout(yaxis=dict(tickformat='+,.0%'), coloraxis=dict(showscale=False))
    fig.update_traces(texttemplate='%{text:+,.0%}', textposition='outside',
                      hovertemplate='<b>%{x}</b><br>%{customdata[0]:+.2f} (%{marker.color:+,.2%})')

    return fig


import altair as alt
from capon.preprocessing import normalize_traces


def plot_history(
    stocks, normalize=None, index="timestamp", value="close", by="symbol", point=False
):
    value_params = {}

    if normalize is not None:
        if normalize == "start":
            normalize = stocks[index].min()
        stocks = normalize_traces(stocks, to=normalize, index=index, by=by)
        value_params = dict(
            axis=dict(format="+%"),
            tooltip=dict(format="+.2%"),
        )

    line_kwargs = {}
    if point:
        line_kwargs = dict(point={"filled": False, "fill": "white"})

    highlight = alt.selection(type="single", on="mouseover", fields=[by], nearest=True)

    base = alt.Chart(stocks).encode(
        x=f"{index}:T",
        y=alt.Y(f"{value}:Q", axis=alt.Axis(**value_params.get("axis", {}))),
        color=f"{by}:N",
    )

    points = (
        base.mark_circle()
        .encode(
            opacity=alt.value(0),
            tooltip=[index, alt.Tooltip(value, **value_params.get("tooltip", {})), by],
        )
        .add_selection(highlight)
    )

    lines = base.mark_line(**line_kwargs).encode(
        # size=alt.condition(~highlight, alt.value(2), alt.value(3))
    )

    chart = points + lines

    if normalize is not None:
        rule = (
            alt.Chart(stocks[stocks[index] >= normalize].nsmallest(1, index))
            .mark_rule(color="black")
            .encode(x=f"{index}:T")
        )

        chart += rule

    return chart.properties(
        width=800,
        height=240,
        title=f"Historical Changes"
        + (f" from {normalize}" if normalize is not None else ""),
    )


if __name__ == "__main__":
    import pandas as pd
    from tqdm.auto import tqdm
    import capon

    tickers = ["BAC", "ALLY", "DFS", "SCHW"]

    history = pd.concat(
        [capon.stock(ticker, range="2y", interval="1mo") for ticker in tqdm(tickers)],
        ignore_index=True,
    )

    chart = plot_history(history)
    chart.display()

    chart = plot_history(history, normalize="start")
    chart.display()

    chart.configure_line(point={"filled": False, "fill": "white"}).display()

    normalize = history["timestamp"].unique()[12].strftime("%Y-%m-%d")

    chart = plot_history(
        history,
        normalize=normalize,
        point=True,
    )
    chart.display()


if __name__ == "__main__":
    from capon.datasets import load_stock_indexes

    tickers = (
        load_stock_indexes()
        .pipe(lambda df: df[df["country_code"] == "US"])
        .symbol.tolist()
    )

    stocks = pd.concat(
        [
            capon.stock(ticker, range="10y", interval="1mo").iloc[:-1]
            for ticker in tqdm(tickers)
        ],
        ignore_index=True,
    ).pipe(lambda df: df[df["timestamp"] >= "2013"])
    stocks

    capon.plot(
        stocks.merge(load_stock_indexes(), on="symbol"), by="name", normalize="start"
    )



import logging
import os

import pandas as pd
from tqdm.auto import tqdm
from joblib import Parallel, delayed

from capon.preprocessing import listify

from .backends.yahoo import stock


logger = logging.getLogger(__name__)


def stock_wrapped(ticker, errors="ignore", *args, **kwargs):
    """Fetch stock history, w/o exception handling.

    Parameters
    ----------
    ticker : str
        Stock ticker.
    errors : {'ignore', 'raise'}, default 'ignore'
        If 'ignore', suppress error and only existing labels are dropped.

    Returns
    -------
    DataFrame
        Stock quotes.

    Raises
    ------
    ValueError
        _description_
    """

    logger.debug(
        f"#{os.getpid()}: stock_wrapped({ticker}, {args}, {', '.join('{0}={1!r}'.format(k, v) for k, v in kwargs.items())}) starts.."
    )

    try:
        return stock(ticker, *args, **kwargs)
    except Exception as e:
        logger.warning(e)
        if errors == "raise":
            raise ValueError(f"symbol '{ticker}' couldn't be retrieved!")
    return None


def stocks(
    symbols,
    range="1d",
    interval=None,
    start=None,
    end=None,
    timestamp_normalizer="auto",
    n_jobs=None,
    *args,
    **kwargs,
):
    symbols = listify(symbols)
    assert len(set(symbols)) == len(symbols), "tickers contains duplicated symbols"

    n_jobs = 1 if n_jobs is None else n_jobs
    if n_jobs != 1:
        quotes_list = Parallel(n_jobs=-1)(
            delayed(stock_wrapped)(
                ticker, range=range, interval=interval, start=start, end=end
            )
            for ticker in tqdm(symbols)
        )
    else:
        quotes_list = [
            stock_wrapped(ticker, range=range, interval=interval, start=start, end=end)
            for ticker in tqdm(symbols)
        ]

    quotes = pd.concat(quotes_list, ignore_index=True)
    # quotes = quotes.dropna(subset=["close", "adjclose"], how="all")
    if timestamp_normalizer == "auto":
        full_days = (
            quotes.groupby(["symbol"])["timestamp"]
            .apply(pd.to_datetime)
            .diff()
            .dt.total_seconds()
            .dropna()
            .pipe(lambda s: (s / (24 * 3600) == (s / (24 * 3600)).astype(int)).all())
        )
        if full_days:
            timestamp_normalizer = "date"

    if timestamp_normalizer == "date":
        quotes = quotes.assign(
            # For multiple timezones
            timestamp=lambda df: pd.to_datetime(df["timestamp"], utc=True)
        ).assign(
            timestamp=lambda df: pd.to_datetime(df["timestamp"].dt.date),
            # timestamp=lambda df: df["timestamp"]
            # .dt.tz_localize(None)
            # .dt.normalize()
        )

    missing_symbols = set(symbols) - set(quotes["symbol"].unique())
    if missing_symbols:
        logger.debug(
            f"{len(missing_symbols):,} symbols were not found ({', '.join(missing_symbols)})"
        )

    return quotes



"""Backend implementation for Yahoo Finance.

"""

import logging
import os
import requests
from datetime import datetime

import pandas as pd
from capon.backends import utils


logger = logging.getLogger(__name__)


ROOT = os.path.dirname(__file__)
valid_range_intervals = {
    "1d": "2m",
    "5d": "15m",
    "1mo": "1h",
    "3mo": "1d",
    "6mo": "1d",
    "1y": "1wk",
    "2y": "1wk",
    "5y": "1mo",
    "10y": "1mo",
    "ytd": "1d",
    "max": "1m",
}


# # 1Y (w)
# url = f'https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?region=US&lang=en-US&includePrePost=false&interval=1wk&range=1y&corsDomain=finance.yahoo.com&.tsrc=finance'
# # 1Y (d)
# url = f'https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?region=US&lang=en-US&includePrePost=false&interval=1d&range=1y&corsDomain=finance.yahoo.com&.tsrc=finance'
# # 1d
# url = f'https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?region=US&lang=en-US&includePrePost=false&interval=2m&range=1d&corsDomain=finance.yahoo.com&.tsrc=finance'


def get_stock(symbol, range="1d", interval=None, start_date=None, end_date=None):
    if range not in valid_range_intervals.keys():
        raise ValueError(
            f"range must be from {list(valid_range_intervals.keys())}, got '{range}'"
        )
    interval = interval or valid_range_intervals[range]
    url = f"https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?region=US&lang=en-US&includePrePost=false&interval={interval}&range={range}&corsDomain=finance.yahoo.com&.tsrc=finance"

    if (start_date is not None) and (end_date is not None):
        start_ts = int(datetime.strptime(start_date, "%Y-%m-%d").timestamp())
        end_ts = int(datetime.strptime(end_date, "%Y-%m-%d").timestamp())
        url = f"https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?region=US&lang=en-US&includePrePost=false&interval={interval}&range={range}&corsDomain=finance.yahoo.com&.tsrc=finance"
        url = (
            f"https://query1.finance.yahoo.com/v8/finance/chart/{symbol}?"
            f"region=US&lang=en-US&includePrePost=false&interval={interval}&corsDomain=finance.yahoo.com&.tsrc=finance"
            f"&period1={start_ts}&period2={end_ts}&interval={interval}"
        )

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:91.0) Gecko/20100101 Firefox/91.0"
    }
    response = requests.get(url, headers=headers)
    logger.info(f"{response.url}.. {response.status_code}")

    jo = response.json()

    if response.status_code != 200:
        logger.info(
            f'Repsonse for "{symbol}" was {response.status_code}:',
            jo["chart"]["error"]["description"],
        )

    return jo


def stock(symbol, range="1d", interval=None, start=None, end=None):
    """Get live & historical stock prices.

    Parameters
    ----------
    symbol : str
        The ticker symbol of the required stock.
    range : {'1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y', '10y', 'ytd', 'max'}, default '1d'
        Specify the timeframe for the required data.
    interval : {None, '1m', '2m', '15m', '1h', '1d', '1wk', '1mo'}, default None
        Specify the time interval for the required data.
        If None, then the default interval of the specified `range` is used.

    Returns
    -------
    DataFrame
        Listing stock price features for all timepoints in the required timeframe. Each row includes the open, high,
        low, and close price for a given timepoint. If relevant, it will also include the adjusted closing price.
    """
    jo = get_stock(
        symbol, range=range, interval=interval, start_date=start, end_date=end
    )

    result = jo["chart"]["result"][0]
    metadata = result["meta"]
    ts = pd.DataFrame(
        {
            "timestamp": pd.to_datetime(result.get("timestamp", []), unit="s")
            .tz_localize("UTC")
            .tz_convert(metadata["gmtoffset"]),
            "symbol": metadata["symbol"],
            "currency": metadata["currency"],
        }
    )
    quote = pd.DataFrame(result["indicators"]["quote"][0])
    adjclose = (
        pd.DataFrame(result["indicators"]["adjclose"][0])
        if "adjclose" in result["indicators"]
        else None
    )

    # len(ts), len(quote), len(adjclose)
    stock = pd.concat([ts, quote, adjclose], axis=1).pipe(
        utils.reindex,
        ["timestamp", "symbol", "currency", "open", "low", "high", "close", "volume"],
    )

    return stock


def history(symbol, start_date, end_date, interval="1mo"):
    start_ts = int(datetime.strptime(start_date, "%Y-%m-%d").timestamp())
    end_ts = int(datetime.strptime(end_date, "%Y-%m-%d").timestamp())

    url = (
        f"https://query1.finance.yahoo.com/v7/finance/download/{symbol}?"
        f"period1={start_ts}&period2={end_ts}&interval={interval}&includeAdjustedClose=true&includePrePost=false"
    )

    # range, interval = "ytd", "1wk"
    # # range, interval = "1d", "2m"
    # range, interval = "1d", "1h"

    # url = (
    #     f"https://query1.finance.yahoo.com/v7/finance/download/{symbol}?"
    #     f"range={range}&interval={interval}&includeAdjustedClose=true&includePrePost=false"
    # )
    logger.info(url)

    history = (
        pd.read_csv(url, parse_dates=["Date"])
        .assign(symbol=symbol)
        .rename(str.lower, axis=1)
        .rename({"date": "timestamp", "adj close": "adjclose"}, axis=1)
        .set_index(["timestamp", "symbol"])
        .reset_index()
    )
    return history


if __name__ == "__main__":
    logging.basicConfig()
    logger.setLevel(logging.INFO)

    stocks1 = history(
        "AAPL", start_date="1980-01-01", end_date="2025-01-01", interval="1wk"
    )
    stocks1

    stocks2 = stock(
        "AAPL", start_date="1980-01-01", end_date="2025-01-01", interval="1d"
    )



"""
curl 'https://query2.finance.yahoo.com/v1/finance/visualization?crumb=21tOl9uwqEm&lang=en-US&region=US&corsDomain=finance.yahoo.com' \
  -H 'authority: query2.finance.yahoo.com' \
  -H 'accept: */*' \
  -H 'accept-language: en-US,en;q=0.9,he;q=0.8' \
  -H 'content-type: application/json' \
  -H 'cookie: B=d3e9e9th7pt8r&b=3&s=ur; A1=d=AQABBBv1fGICEAQrPN2cJPkveGzSC08uudEFEgEBBgGwh2JtY15Ub2UB_eMBAAcIG_V8Yk8uudE&S=AQAAAl39dQ4SdhV06WCVAtJUlb8; A3=d=AQABBBv1fGICEAQrPN2cJPkveGzSC08uudEFEgEBBgGwh2JtY15Ub2UB_eMBAAcIG_V8Yk8uudE&S=AQAAAl39dQ4SdhV06WCVAtJUlb8; GUC=AQEBBgFih7BjbUIeGAR1; thamba=2; A1S=d=AQABBBv1fGICEAQrPN2cJPkveGzSC08uudEFEgEBBgGwh2JtY15Ub2UB_eMBAAcIG_V8Yk8uudE&S=AQAAAl39dQ4SdhV06WCVAtJUlb8&j=WORLD; PRF=t%3DALLY%252BAAPL%252B%255EDJI%252BAAIC%252BREI-UN.TO%252BREML%252BEVGN.TA%252BORA.TA%252BORA%252B%255EGSPC%252BBTC-USD%252BMPNGY%252B3690.HK%252BSHEL%252BMPLX; cmp=t=1657277721&j=0&u=1---' \
  -H 'origin: https://finance.yahoo.com' \
  -H 'referer: https://finance.yahoo.com/calendar/earnings?from=2022-06-05&to=2022-06-11&day=2022-06-05' \
  -H 'sec-ch-ua: " Not A;Brand";v="99", "Chromium";v="102", "Google Chrome";v="102"' \
  -H 'sec-ch-ua-mobile: ?0' \
  -H 'sec-ch-ua-platform: "macOS"' \
  -H 'sec-fetch-dest: empty' \
  -H 'sec-fetch-mode: cors' \
  -H 'sec-fetch-site: same-site' \
  -H 'user-agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36' \
  --data-raw '{"sortType":"ASC","entityIdType":"earnings","sortField":"companyshortname","includeFields":["ticker","companyshortname","startdatetime","startdatetimetype","epsestimate","epsactual","epssurprisepct","timeZoneShortName","gmtOffsetMilliSeconds"],"query":{"operator":"and","operands":[{"operator":"gte","operands":["startdatetime","2022-06-06"]},{"operator":"lt","operands":["startdatetime","2022-06-07"]},{"operator":"eq","operands":["region","us"]}]},"offset":0,"size":100}' \
  --compressed
"""


import json
from pprint import pprint
from urllib.request import urlopen
from urllib.parse import urlencode

def parse():
    host   = 'https://query1.finance.yahoo.com'
    #host   = 'https://query2.finance.yahoo.com'  # try if above doesn't work
    path   = '/v10/finance/quoteSummary/%s' % 'ADP'
    params = {
        'formatted' : 'true',
        #'crumb'     : 'ILlIC9tOoXt',
        'lang'      : 'en-US',
        'region'    : 'US',
        'modules'   : 'earningsTrend',
        'domain'    : 'finance.yahoo.com'
    }

    response = urlopen('{}{}?{}'.format(host, path, urlencode(params)))
    data = json.loads(response.read().decode())

    pprint(data)

if __name__ == '__main__':
    parse()


import requests
from bs4 import BeautifulSoup


url = "https://dividendhistory.org/payout/BAC/"
# response = requests.get(url)
# response.content


# soup = BeautifulSoup(response.text, "html.parser")
# soup.find(id="dividend_table")


import pandas as pd

dividends = pd.read_html(url)[1]
for ts in ["Ex-Dividend Date", "Payout Date"]:
    dividends[ts] = pd.to_datetime(dividends[ts])
dividends["Cash Amount"] = (
    dividends["Cash Amount"].str.lstrip("$").str.rstrip("**").astype(float)
)



import pandas as pd

from capon.backends.utils import camel_case_keys

urls = {
    'screener': 'https://www.otcmarkets.com/research/stock-screener/api/downloadCSV',
}


def screener():
    url = urls['screener']
    metadata = pd.read_csv(url).pipe(camel_case_keys)

    return metadata


if False:
    metadata = screener()
    metadata.columns


import logging
from .wrappers import stock, stocks
from .backends.nasdaq import metadata


logger = logging.getLogger(__name__)


try:
    from .visualization import template
except ModuleNotFoundError as e:
    logger.warning(e)

try:
    # from .visualization import plot_history as plot
    from .visualization.altairplotter import plot_history as plot
except ModuleNotFoundError as e:
    logger.warning(e)

    def plot(*args, **kwargs):
        raise ImportError(
            "Missing optional dependency 'altair'. Use pip or conda to install it."
        )



import logging
import os


RESOURCES_ROOT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '../resources'))
logging.info(f'RESOURCES_ROOT_PATH = "{RESOURCES_ROOT_PATH}"')


def get_resource_path(path):
    return os.path.normpath(f'{RESOURCES_ROOT_PATH}/{path}')


import numpy as np
import pandas as pd


def listify(x):
    x = [x] if np.ndim(x) == 0 else x
    return x


def normalize_traces(
    traces,
    to="start",
    index="timestamp",
    values=None,
    by="symbol",
    method="independent",
    baseline=1,
):
    if to == "start":
        if method == "independent":
            to = traces[index].min()
        elif method == "common":
            raise NotImplementedError
        else:
            raise ValueError(f'unknown method "{method}"')

    if values is None:
        values = traces.select_dtypes(np.number).columns.tolist()
    else:
        values = listify(values)

    normalized = (
        traces.groupby(by)
        .apply(
            lambda g: g[values] / g[g[index] >= to].nsmallest(1, index).iloc[0][values]
            - baseline,
        )
        .reset_index(by, drop=True)
        .astype(float)
    )

    assert traces.index.nunique() == len(traces), "traces index must be unique"
    return traces[[index, by]].join(normalized, validate="1:1")


def shift(g, index="timestamp", value="close", days=(-1, -7), fillna=None):
    assert g[index].nunique() == len(g), "{index} should be unique"
    assert all(g[[index, value]].notna())

    days = listify(days)
    s = g.set_index([index], drop=False)[[index, value]]

    windows = pd.concat(
        [s]
        + [
            s.shift(k, freq=pd.DateOffset(days=-1)).add_suffix(
                f"_{'prev' if k<0 else 'next'}{abs(k)}days"
            )
            for k in days
        ],
        axis=1,
    ).sort_index()

    if fillna is not None:
        fillna = {
            "bfill": lambda df: df.bfill(),
            "ffill": lambda df: df.ffill(),
        }[fillna]

        windows = pd.concat([s, windows.drop(s.columns, axis=1).pipe(fillna)], axis=1)

    windows = g.merge(
        windows.dropna(subset=[index, value]).reset_index(drop=True),
        on=[index, value],
        how="left",
        validate="one_to_one",
    )
    windows.index = g.index  # hack since merge cleans index

    return windows


if __name__ == "__main__":
    history = pd.DataFrame(
        [
            ["2014-01-27 01:00:00-04:00", 22.410451889038086],
            ["2014-02-03 01:00:00-04:00", 22.833295822143555],
            ["2014-02-10 01:00:00-04:00", 22.833295822143555],
            ["2014-02-17 01:00:00-04:00", 22.833295822143555],
            ["2014-02-24 01:00:00-04:00", 22.833295822143555],
            ["2014-03-03 01:00:00-04:00", 22.833295822143555],
            ["2014-03-10 00:00:00-04:00", 23.890390396118164],
            ["2014-03-17 00:00:00-04:00", 23.890390396118164],
            ["2014-03-24 00:00:00-04:00", 24.38370132446289],
            ["2014-03-31 00:00:00-04:00", 24.38370132446289],
            ["2014-04-07 00:00:00-04:00", 21.147573471069336],
            ["2014-04-14 00:00:00-04:00", 20.867938995361328],
            ["2014-04-21 00:00:00-04:00", 21.313613891601562],
            ["2014-04-28 00:00:00-04:00", 21.016498565673828],
        ],
        columns=["timestamp", "close"],
    ).assign(timestamp=lambda df: pd.to_datetime(df["timestamp"]))

    shift(history, days=7)
    shift(history, days=30, fillna="bfill").assign(
        d=lambda df: df["timestamp_next30days"] - df["timestamp"]
    )



import logging
import datetime
from enum import Enum
from dataclasses import dataclass, field

import pandas as pd
from tqdm.auto import tqdm

import capon


logger = logging.getLogger(__name__)
default_indexes = ["^GSPC", "^DJI", "^IXIC", "^RUT"]
local_tz = datetime.datetime.now().astimezone().tzinfo


@dataclass
class Lot:
    """
    Attributes
    ----------
    timestamp : str
        Trade date
    symbol : str
        The stock symbol
    quantity : int
        The amount of stocks
    price: float
        Cost per share
    """

    timestamp: str
    symbol: str
    quantity: int
    price_buy: float
    active: bool = 1

    def __post_init__(self):
        self.timestamp = pd.to_datetime(self.timestamp)


# pd.DataFrame([Lot("2020-01-01", "A", 100, 1.00)])
# pd.DataFrame([Lot(pd.to_datetime("2020-01-01"), "A", 100, 1.00)])


class OrderType(Enum):
    BUY = 1
    SELL = 2


@dataclass
class Order:
    """Base class for Buy/Sell orders."""

    timestamp: str
    symbol: str
    quantity: int
    price: float
    type: OrderType

    def __post_init__(self):
        self.timestamp = pd.to_datetime(self.timestamp)


# print(Order("2020-01-01", "AAA", 10, 1.23, OrderType.BUY).type)


@dataclass
class BuyOrder(Order):
    type: OrderType = field(default=OrderType.BUY, init=False)


@dataclass
class SellOrder(Order):
    type: OrderType = field(default=OrderType.SELL, init=False)


# pd.DataFrame([
#     BuyOrder("2020-01-01", "AAA", 10, 1.23),
#     SellOrder("2020-01-01", "AAA", 10, 1.23),
# ])


class Portfolio:
    def __init__(self, lots=None):
        if lots is None:
            trades = pd.DataFrame()
        else:
            trades = (
                pd.DataFrame(lots)
                .assign(
                    timestamp_buy=lambda df: pd.to_datetime(df["timestamp"]),
                )
                .drop("timestamp", axis=1)
                .pipe(Portfolio.enrich)
            )
        self.trades = trades

    def status(self, history=None, timestamp_normalizer="date"):
        performance = self.performance(
            history=history, timestamp_normalizer=timestamp_normalizer
        )

        latest = (
            performance.groupby(["lot_index", "symbol"])
            .apply(
                lambda g: g.sort_values("timestamp")
                .assign(
                    timestamp=lambda df: df["timestamp"].where(
                        df["active"] == 1, df["timestamp_sell"]
                    ),
                    price=lambda df: df["price"].where(
                        df["active"] == 1, df["price_sell"]
                    ),
                )
                .assign(
                    value=lambda df: df["price"] * df["quantity"],
                    gain=lambda df: df["value"] - df["cost"],
                    gain_pct=lambda df: df["value"] / df["cost"] - 1,
                )
                .drop(["timestamp_sell", "price_sell"], axis=1)
                .assign(daily_pct=lambda df: df["price"].pct_change())
                .tail(1)
            )
            .reset_index(drop=True)
        )
        return latest

    def performance(self, history=None, timestamp_normalizer="date", end_time=None):
        price = "adjclose"

        if timestamp_normalizer == "date":
            timestamp_normalizer = lambda ts: pd.to_datetime(ts.dt.date)
        else:
            raise NotImplementedError

        if history is None:
            history = self.fetch_history()
        # @todo, verify coverage of lots

        performance = (
            self.trades.assign(
                timestamp_buy=lambda df: timestamp_normalizer(df["timestamp_buy"])
            )
            .merge(
                history[["symbol", "timestamp", price]]
                .rename({price: "price"}, axis=1)
                .assign(timestamp=lambda df: timestamp_normalizer(df["timestamp"])),
                on="symbol",
                how="left",
            )
            .assign(
                value=lambda df: df["price"] * df["quantity"],
                gain=lambda df: df["value"] - df["cost"],
                gain_pct=lambda df: df["value"] / df["cost"] - 1,
            )
            .pipe(lambda df: df[df["timestamp_buy"] <= df["timestamp"]])
            .assign(
                active=lambda df: df["timestamp_sell"].isna()
                | (df["timestamp"] <= df["timestamp_sell"])
            )
            .astype({"active": int})
            #             .assign(timestamp_last=lambda df: df[["timestamp_sell", "timestamp"]].min(axis=1))
            #             .pipe(lambda df: df[df["timestamp"]<=df["timestamp_last"]])
            #             .drop(["timestamp_last"], axis=1)
        )

        if end_time is not None:
            performance = performance.pipe(lambda df: df[df["timestamp"] <= end_time])

        return performance

    def buy(self):
        raise NotImplementedError

    def sell(self, timestamp, symbol, quantity, price):
        assert quantity > 0

        available_lots = self.trades.pipe(lambda df: df[df["symbol"] == symbol])
        available_quantity = available_lots["quantity"].sum()
        if available_quantity < quantity:
            raise ValueError(
                f"Insufficient stocks, {quantity} were ordered to be sold, but only {available_quantity} are available."
            )

        if len(available_lots) > 1:
            raise NotImplementedError

        self.trades = (
            pd.concat(
                [
                    self.trades.assign(
                        fifo_sell_quantity=lambda df: (df["symbol"] == symbol)
                        * quantity,
                        quantity=lambda df: df["quantity"] - df["fifo_sell_quantity"],
                    )
                    .drop("fifo_sell_quantity", axis=1)
                    .pipe(lambda df: df[~(df["active"] & (df["quantity"] == 0))]),
                    available_lots.iloc[:1].assign(
                        quantity=quantity,
                        active=0,
                        timestamp_sell=timestamp,
                        price_sell=price,
                    ),
                ],
                ignore_index=True,
            )
            .assign(timestamp_sell=lambda df: pd.to_datetime(df["timestamp_sell"]))
            .pipe(Portfolio.enrich)
        )

    def fetch_history(self, symbols=None, start_time=None, end_time=None, tz=None):
        if symbols is None:
            symbols = self.trades["symbol"].unique().tolist()
        if start_time is None:
            start_time = str(self.trades["timestamp_buy"].dt.date.min())
        if end_time is None:
            end_time = str(pd.Timestamp.now().date() + pd.Timedelta(days=1))  # tomorrow

        history = pd.concat(
            [
                capon.stock(
                    symbol, start=start_time, end=end_time, interval="1d"
                ).dropna()
                for symbol in tqdm(symbols)
            ],
            ignore_index=True,
        )

        if tz is not False:
            tz = local_tz if tz is None else tz
            logger.debug(f"Setting timezone to {tz}..")

            history = history.assign(
                timestamp=lambda df: pd.to_datetime(
                    df["timestamp"], utc=True
                ).dt.tz_convert(tz)
            )

        return history

    @staticmethod
    def enrich(trades):
        trades = (
            trades.assign(
                cost=lambda df: df["price_buy"] * df["quantity"],
            )
            .sort_values(["timestamp_buy", "symbol", "active"], ignore_index=True)
            .assign(lot_index=lambda df: range(len(df)))
        )

        if "timestamp_sell" not in trades:
            assert "price_sell" not in trades
            trades = trades.assign(
                timestamp_sell=lambda df: pd.to_datetime([None] * len(df)),
                price_sell=None,
            )

        return trades

    @staticmethod
    def from_lots(lots):
        return Portfolio(lots)

    @staticmethod
    def from_orders(orders, init_lots=None):
        portfolio = Portfolio(init_lots)

        for order in orders:
            if order.type == OrderType.BUY:
                portfolio.buy(**order.__dict__)
            elif order.type == OrderType.SELL:
                portfolio.sell(**order.__dict__)
            else:
                raise ValueError(order)

        return portfolio



import logging
import glob
import re

import pandas as pd

from .common import get_resource_path


stock_indexes = pd.DataFrame([
    ['^GSPC',     'S&P 500',      'S&P 500 Index (SPX)', 'US', 'New York'],
    ['^DJI',      'Dow 30',       'Dow Jones Industrial Average (DJIA)', 'US', 'New York'],
    ['^IXIC',     'Nasdaq',       'NASDAQ Composite Index (COMP)', 'US', 'New York'],
    ['^RUT',      'Russell 2000', 'Russell 2000 Index (RUT)', 'US', 'New York'],
    ['^GDAXI',    'DAX',          'DAX Pefromance Index (DAX)', 'Germany', 'Frankfurt'],
    ['^FTSE',     'FTSE 100',     'The Financial Times Stock Exchange 100 Index (UKX)', 'UK', 'London'],
    ['^N225',     'Nikkei 225',   'NIKKEI 225 Index (NIK)', 'Japan', 'Tokyo'],
    ['^TA125.TA', 'TA-125',       'Tel Aviv 125 Index (TA100)', 'Israel', 'Tel Aviv'],
    ['^BVSP',     'IBOVESPA',     '', 'Brazil', 'Sao Paolo'],
], columns=['symbol', 'name', 'description', 'country_code', 'city'])


def load_stock_indexes():
    """Major worldwide stock indexes

    Returns
    -------
    DataFrame
        Symbols and description for several worldwide stock indexes.

    References
    -------
    See https://finance.yahoo.com/world-indices/, or descriptions in https://www.marketwatch.com/investing/index/djia.

    """
    return stock_indexes.copy()


def load_metadata():
    metadata_filename = sorted(glob.glob(get_resource_path('metadata/stocks.metadata.*.gz')),
                               key=lambda s: parse_metadata_filename(s)['timestamp'])[-1]
    logging.info(f'Loading metadata from "{metadata_filename}"..')

    return pd.read_csv(metadata_filename)


def parse_metadata_filename(filename):
    m = re.match(r'.*stocks.metadata.x(?P<size>\d+)\.\((?P<timestamp>[0-9\.]+)\)\.csv.gz', filename)
    return m.groupdict()





import plotly.express as px

from capon.preprocessing import normalize_traces


def plot_history(stocks, anchor, title=""):
    """Plot historical prices of stocks

    Parameters
    ----------
    stocks
    anchor
    title

    Returns
    -------

    """

    stocks_ts_normed = normalize_traces(stocks, to=anchor)

    fig = px.line(
        stocks_ts_normed,
        x="timestamp",
        y="adjclose",
        color="symbol",
        title=f"{title}: Historical changes from {anchor}",
    )
    fig.update_layout(yaxis_tickformat="+%")
    fig.update_layout(
        shapes=[
            dict(
                type="line",
                yref="paper",
                y0=0,
                y1=1,
                xref="x",
                x0=anchor,
                x1=anchor,
                line=dict(width=1.5, dash="dash", color="rgba(0,0,0,0.5)"),
            )
        ]
    )
    # fig.update_xaxes(rangeslider_visible=True)
    # fig.update_layout(hovermode="x")

    return fig


if __name__ == "__main__":
    import capon

    stocks = capon.stock("TSLA", range="1y")
    fig = plot_history(stocks, anchor=stocks["timestamp"].median())
    fig.show()



from setuptools import setup, find_packages


with open("README.md", "r") as readme_file:
    readme = readme_file.read()


base_packages = [
    "requests>=2.31.0",
    "numpy>=1.17.3",
    "pandas>=1.0.3",
    "tqdm>=4.66.3",
    "joblib",
]
plotly_packages = ["plotly>=4.7.1"]
altair_packages = ["altair"]
test_packages = ["pytest", "ipython", "tox"]
docs_packages = ["black>=24.3.0"]
dev_packages = (
    ["notebook", "matplotlib", "ipywidgets", "seaborn", "themes"]
    + ["scikit-learn>=1.5.0"]
    + altair_packages
    + docs_packages
    + test_packages
)


setup(
    name="capon",
    version="0.0.92",
    author="Eyal Gal",
    author_email="eyalgl@gmail.com",
    description="Capital Market in Python",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords=[
        "capital markets",
        "stocks",
        "stock market",
        "finance",
        "dataset",
        "portfolio",
        "dashboard",
        "yahoo finance",
    ],
    url="https://github.com/gialdetti/capon/",
    packages=find_packages(),
    install_requires=base_packages,
    extras_require={
        "test": test_packages,
        "docs": docs_packages,
        "dev": dev_packages,
    },
    include_package_data=True,
    # package_data={'datasets': ['capon/resources/*']},
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



import logging
import pickle
import os
import argparse
import multiprocessing as mp

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

import capon
from capon.datasets.common import get_resource_path
from capon import datasets
from capon import backends


# logging.getLogger("capon").setLevel(logging.INFO)

exchange_name = "nasdaq"
# exchange_name = "otc"


def load_config(path, create=False, nasdaq=True, otc=False):
    if create:
        indexes_symbols = datasets.load_stock_indexes()["symbol"].tolist()
        exchanges = {
            "indexes": indexes_symbols,
        }

        if nasdaq:
            metadata = datasets.load_metadata()
            screener_metadata = backends.nasdaq.screener()
            print(f"{len(screener_metadata):,} stocks were screened.")

            nasdaq_symbols = np.unique(screener_metadata["symbol"]).tolist()
            print(f"{len(nasdaq_symbols):,} stocks tickers.")
            # nasdaq_symbols = list(
            #     set.union(set(metadata["symbol"]), set(screener_metadata["symbol"]))
            # )

            if True:  # Enrich with SP-500
                sp500_history = pd.read_csv(
                    "/Users/eyalgal/Library/CloudStorage/Dropbox/Work/fusion/code/fpai/markets/resources/sp500_component_history.2022-12-22.csv",
                    index_col="date",
                )
                nasdaq_symbols = sorted(
                    pd.DataFrame(
                        {
                            "symbol_raw": np.unique(
                                # nasdaq_symbols +
                                sp500_history.columns.tolist()
                            )
                        }
                    )
                    .assign(
                        symbol=lambda df: df["symbol_raw"]
                        .str.replace(".", "-")
                        .str.replace("/", "-")
                        .str.replace("^", "-P")
                        .str.strip()
                    )["symbol"]
                    .unique()
                )

                print(
                    f"{len(nasdaq_symbols):,} stocks tickers after enriching with historical SP500 ({sp500_history.shape[1]:,} tickers)."
                )

            exchanges["nasdaq"] = nasdaq_symbols

        if otc:
            otc_metadata = backends.otcmarkets.screener()
            otc_symbols = list(otc_metadata["symbol"])

            exchanges["otc"] = otc_symbols

        os.makedirs(path, exist_ok=True)

        exchanges_filename = f"{path}/exchange.pickle"
        print(f"Saving {exchanges_filename}..")
        with open(exchanges_filename, "wb") as handle:
            pickle.dump(exchanges, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        exchanges_filename = f"{path}/exchange.pickle"
        print(f"Loading {exchanges_filename}..")
        with open(exchanges_filename, "rb") as handle:
            exchanges = pickle.load(handle)

    print("exchanges:", {k: len(v) for k, v in exchanges.items()})
    return exchanges


def fetch(ticker, range="10y"):
    # ticker = ticker.replace(".", "-").replace("/", "-")
    try:
        # return capon.stock(ticker, range=range, interval="1d")
        # return backends.yahoo.history(
        #     ticker, start_date="1995-01-01", end_date="2025-01-01", interval="1d"
        # )

        start_date, end_date = "1980-01-01", "2025-01-01"
        # start_date, end_date = "1980-01-01", "2022-10-07"

        return capon.stock(ticker, start=start_date, end=end_date, interval="1d")
    except:
        return None


def get_history(required_symbols, path):
    assert len(set(required_symbols)) == len(
        required_symbols
    ), "required_symbols contains duplicated symbols"

    pool = mp.Pool(mp.cpu_count())
    stocks = pd.concat(
        tqdm(pool.imap(fetch, required_symbols), total=len(required_symbols)),
        ignore_index=True,
    )
    pool.close()

    stocks = stocks.dropna(subset=["close", "adjclose"], how="all").assign(
        timestamp=lambda df: pd.to_datetime(df["timestamp"], utc=True).dt.date
    )

    stocks_filename = f"{path}/" + "capon.history.{}.{}-{}.x{:05}.({}).csv.gz".format(
        exchange_name,
        stocks["timestamp"].min().strftime("%y%m%d"),
        stocks["timestamp"].max().strftime("%y%m%d"),
        stocks["symbol"].nunique(),
        pd.Timestamp.now().strftime("%y%m%d.%H%M%S"),
    )

    print(f"Saving {stocks_filename}..")
    stocks.to_csv(stocks_filename, index=False, compression="gzip")

    missing_symbols = set(required_symbols) - set(stocks["symbol"].unique())
    print(f"{len(missing_symbols)} symbols were not found: {missing_symbols}")

    pd.DataFrame({"symbol": required_symbols}).assign(
        downloaded=lambda df: df["symbol"].isin(stocks["symbol"])
    ).to_csv(stocks_filename.replace(".csv.gz", ".LOG.csv"), index=False)


def get_metadata(required_symbols, path):
    pool = mp.Pool(mp.cpu_count())
    metadata = pd.concat(
        tqdm(
            pool.imap(backends.nasdaq.metadata, required_symbols),
            total=len(required_symbols),
        ),
        axis=1,
        ignore_index=True,
    ).T
    pool.close()

    metadata = metadata.drop("module_title", axis=1).dropna(how="all")
    metadata_filename = f'{path}/capon.metadata.{exchange_name}.x{metadata["symbol"].nunique():05}.({pd.Timestamp.now().strftime("%y%m%d.%H%M%S")}).csv.gz'

    print(f"Saving {metadata_filename}..")
    metadata.to_csv(metadata_filename, index=False, compression="gzip")


def get_dividends(required_symbols, path):
    pass


def download(args):
    dataset_id = args.id
    do_create_config = False

    if dataset_id is None:
        dataset_id = pd.Timestamp.now().strftime("%y%m%d.%H%M%S")
        print(f'new dataset_id: "{dataset_id}"')
        do_create_config = True

    # path = get_resource_path(f"../../outputs/capon-data/__raw_{dataset_id}")
    path = os.path.expanduser(
        f"~/Downloads/fp-datasets/capon/capon-data/__raw_{dataset_id}"
    )
    print(f"path: {path}")

    if True:
        print("\nConfiguration..")
        exchanges = load_config(path, create=do_create_config)

    required_symbols = list(np.unique(exchanges[exchange_name] + exchanges["indexes"]))
    if args.debug:
        required_symbols = required_symbols[:20]
    print(f"{len(required_symbols):,} required symbols")

    if True:
        print("\nDownloading history..")
        get_history(required_symbols, path=path)

    if True:
        print("\nDownloading metadata..")
        get_metadata(required_symbols, path=path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--id",
        default=None,
        type=str,
        help="Dataset id. If None, a new will be created.",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug mode (for a fast running time)."
    )
    args = parser.parse_args()

    if args.debug:
        print(f"args: {args}")

    download(args)


if False:
    stocks = pd.read_csv(
        "~/Dropbox/PhD/Code/capon/outputs/metadata/raw_220705.101823/capon.history.nasdaq.120702-220704.x07859.(220705.105016).csv.gz"
    )
    metadata = pd.read_csv(
        "~/Dropbox/PhD/Code/capon/outputs/metadata/raw_220705.101823/capon.metadata.nasdaq.x07878.csv.gz"
    )

    set(metadata["symbol"]) - set(stocks["symbol"])
    """
    {'ASCBR',
    ...
    'EAI',
    'ECC           ',
    'ECOM          ',
    'EMP',
    """

    import glob

    last_dataset_folder = sorted(
        glob.glob(get_resource_path(f"../../outputs/capon-data/*"))
    )[-1]
    last_dataset_history = glob.glob(f"{last_dataset_folder}/*.history.*.csv.gz")[0]

    print(f"Loading {last_dataset_history.split('/')[-1]}")
    stocks = pd.read_csv(last_dataset_history)
    stocks["timestamp"].agg(["min", "max"])
    stocks.groupby(["symbol"])["timestamp"].agg(["min", "max", "count"])

if False:
    start_date, end_date = "1980-01-01", "2022-10-07"
    capon.stock("AMZN", start_date=start_date, end_date=end_date, interval="1d")



import pandas as pd

from capon.backends.utils import get_json, camel_case_keys

urls = {
    "screener": "https://api.nasdaq.com/api/screener/stocks?tableonly=true&exchange={exchange}&download=true",
    "profile": "https://api.nasdaq.com/api/company/{symbol}/company-profile",
    "info": "https://api.nasdaq.com/api/quote/{symbol}/info?assetclass=stocks",
    "summary": "https://api.nasdaq.com/api/quote/{symbol}/summary?assetclass=stocks",
    "chart": "https://api.nasdaq.com/api/quote/{symbol}/chart?assetclass=stocks&fromdate={from_date}&todate={to_date}",
}

headers = {
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36"
}


def screener(exchange=""):
    """Full stock metadata allowing further screening.

    Parameters
    ----------
    exchange : {'', 'NASDAQ', 'NYSE', 'AMEX'}, default ''
        The stock exchange of the required list. By default will return symbols from all stock exchanges.


    Returns
    -------
    DataFrame
        Listing of all stocks and their basic metadata (e.g., market_cap, industry, sector).
    """

    url = urls["screener"].format(exchange=exchange)
    jo = get_json(url, headers)
    metadata = pd.DataFrame(jo["data"]["rows"]).pipe(camel_case_keys)

    return metadata


def metadata(*args, **kwargs):
    return profile(*args, **kwargs)


def profile(symbol):
    url = urls["profile"].format(symbol=symbol)
    jo = get_json(url, headers)

    if ("data" in jo) and (jo["data"] is not None):
        profile = (
            pd.DataFrame(jo["data"]).loc["value"].rename(symbol).pipe(camel_case_keys)
        )
    else:
        profile = pd.Series(None, name=symbol, dtype="object")

    return profile


def info(symbol):
    url = urls["info"].format(symbol=symbol)
    jo = get_json(url, headers)
    info = pd.Series(jo["data"], name=symbol).pipe(camel_case_keys)

    return info


def summary(symbol):
    url = urls["summary"].format(symbol=symbol)
    jo = get_json(url, headers)

    if jo["data"] is not None:
        summary = (
            pd.DataFrame(jo["data"]["summaryData"])
            .loc["value"]
            .rename(symbol)
            .pipe(camel_case_keys)
        )
    else:
        summary = pd.Series(None, name=symbol, dtype="object")

    return summary


def chart(symbol, from_date="", to_date=""):
    url = urls["chart"].format(symbol=symbol, from_date=from_date, to_date=to_date)
    jo = get_json(url, headers)

    chart_metadata = pd.Series(jo["data"], name=symbol).pipe(camel_case_keys)
    if (jo["data"] is not None) and (jo["data"]["chart"] is not None):
        chart_data = (
            pd.DataFrame(chart_metadata["chart"])["z"]
            .apply(pd.Series)
            .assign(timestamp=lambda df: pd.to_datetime(df["dateTime"], dayfirst=True))
            .drop("dateTime", axis=1)
        )
    else:
        chart_data = None

    return chart_data, chart_metadata


if False:
    symbol = "AMZN"
    symbol = "QUBT"

    profile(symbol)
    info(symbol)
    summary(symbol)
    chart(symbol)

    chart_data, chart_metadata = chart(
        symbol, from_date="2019-12-24", to_date="2020-12-31"
    )

    metadata = screener()
    metadata.shape
    metadata.columns



import plotly.graph_objects as go
import plotly.io as pio
import plotly.express as px


pio.templates["capon"] = go.layout.Template(
    layout=go.Layout(
        font=dict(color="#887e7d"),
        paper_bgcolor="#fff1e5",
        plot_bgcolor="#fff1e5",
        title=dict(
            font=dict(size=22, color="black", family="Palatino"), xanchor="left", x=0.05
        ),
        colorway=px.colors.qualitative.T10,
        colorscale={
            "sequential": "Geyser_r",
            "sequentialminus": "Geyser_r",
        },
    ),
    data_scatter=[go.Scatter(line=dict(width=2.5))],
)

# https://plotly.com/python/builtin-colorscales/
# https://plotly.com/python/discrete-color/


# from plotly.offline import plot
# fig = px.line(x=[1,2,3], y=[1,3,2], title='Test', template='capon')
# plot(fig)
